function [firstDetection, result, state_up, state_down] = NPCUSUM_test(dataset, NPCUSUM_parameters)
%NPCUSUM_TEST
%   Component-wise NPCUSUM change detection test based on
% "Detection of intrusions in information systems by sequential change-point methods"
% Tartakovsky et al., Statistical Methodology 3 (2006)
%
% Input:
%       - dataset: dataset vector to be analyzed.
%           For Multidimensional data, each row of the dataset
%           should correspond to a component, thus
%           size(DataSet , 1) = sample dimension, (d in [1])
%           size(DataSet , 2) = number of samples of the dataset
%       - NPCUSUM_parameters: parameters configuration for the
%          NPCUSUM change detection test
%
% Output:
%       - firstDetection: first sample with detection
%       - result: detection for each sample of the dataset
%       - state: value of the figure of merit for each sample of the
%           dataset
%
% Manuel Roveri
% Politecnico di Milano
% November 2010
% roveri@elet.polimi.it

%% Change Detection Test Setup
[number_of_component, number_of_samples] = size(dataset);
training_set_length = NPCUSUM_parameters.training_set_length;
w = NPCUSUM_parameters.w;
result = zeros(number_of_component, number_of_samples);
state = zeros(number_of_component, number_of_samples);
epsilon = NPCUSUM_parameters.epsilon;
h = NPCUSUM_parameters.h;
firstDetection = zeros(number_of_component, 1);

%% The test is so each feature is managed separately
for curr_component = 1 : number_of_component

    % 1) Estimation of mu
    mu = mean(dataset(curr_component, 1 : training_set_length));
      
    % 2) Start analyzing the data
    S_up = 0;
    S_down = 0;
    for curr_sample = training_set_length + 1 : number_of_samples

        % 3.a) Compute c
        switch lower(NPCUSUM_parameters.c.current)
            case 'fixed'
                c = NPCUSUM_parameters.c.fixed.value;
            case 'sliding_window'
                if (curr_sample - NPCUSUM_parameters.c.fixed.window_size) <= 0
                    disp('NPCUSUM cdt - ERROR - Fixed window too big!!!')
                    return;
                end
                c = epsilon * mean( dataset(curr_component, ...
                    curr_sample - NPCUSUM_parameters.c.fixed.window_size + 1 : curr_sample));
            otherwise
                disp('NPCUSUM cdt - ERROR - Unknown method to compute c!!!')
                return;
        end


        % 3.b) Update the figure of merit
 
        % VERSIONE VARIAZIONI SIA SALITA CHE DISCESA
        S_up = max (0, S_up + w * (dataset(curr_component, curr_sample) - mu - c));        
        S_down = max (0, S_down + w * (-dataset(curr_component, curr_sample) + mu - c));        
       
        % 3.c) Update the state matrix
        state_up (curr_component, curr_sample) = S_up;
        state_down (curr_component, curr_sample) = S_down;        
        
        % 3.d) Update the result matrix
        if (S_up > h) || (S_down > h)
            result (curr_component, curr_sample) = 1;
            if firstDetection (curr_component) == 0
                firstDetection (curr_component) = curr_sample;
            end
            
            break

        end


    end
end

end
