function [tsEnd, tsInit, confirmedDetections, tsLengths, originalDetections] = f_NPCUSUM_MultiChange_MultiStage(DataSet, params)
%
% function [tsEnd, tsInit, confirmedDetections, tsLengths, originalDetections] = f_NPCUSUM_MultiChange_MultiStage(DataSet, params)
%
% input description
%
% DataSet              a row vector containing the sequence of scalar data to be analyzed (data are analyzed in an online, sequential way)
%
% params 		      the parameter structure for the test. This can be generated by the function define_ICI_test_parameters()
%                            that sets the default parameters
%
% output description
%
% tsEnd                                  vector of the detected change-time instants: i.e max([\hat{T}, T_{ref} + params.MinimumTS_Size]). 
%                                               when detections are not validated by the second layer this vector
%                                               contains repeated values.  Note that the output of the first layer CDT
%                                               are in the vector confirmedDetections (third argument), and values of
%                                               tsEnd are often delayed to gather at least params.MinimumTS_Size
%                                               observations after the prospective change.
%                                               tsInit(1) = 0 as NP CUSUM requires no training sequence
% tsInit                                   vector of refined change-time instants (i.e T_{ref}). 
%                                               when detections are not validated by the second layer this vector
%                                               contains repeated values.
%                                               tsInit(1) = 1 as it corresponds to the initial training set
% confirmedDetections       vector of the estimated change time instant from the first layer implementing 
%                                               algorithm in [1]. Thus, only CDT detections validated by the Hotelling T^2 
%                                               hypothesis test are here reported. 
%                                              IMPORTANT REMARK: we do not impose any minimum distance between tsInit and
%                                               confirmedDetections. However, a minimum number of observations is required 
%                                               between tsInit and tsEnd, to validate the change and possibly reconfigure the CDT.
%                                               When CDT have to be assessed on a sequence of changes, then tsEnd
%                                               should be taken into account.
% tsLengths                         vector lengts of each trainig set (i.e for each change \hat{T} - T_{ref})
% originalDetections          detection of ICI-based CDT without validation: it corresponds to the detections raised by
%                                            the ICI-based CDT in [3]. When there are no false positives,
%                                            originalDetections and confirmedDetections do coincide
%
% 
% Giacomo Boracchi
% Politecnico di Milano
% November 2013     
% giacomo.boracchi@polimi.it
%
% REFERNCES
%
% [1] Component-wise NPCUSUM change detection test based on
% "Detection of intrusions in information systems by sequential change-point methods"
% Tartakovsky et al., Statistical Methodology 3 (2006)
%
% [2] Cesare Alippi, Giacomo Boracchi, Manuel Roveri,
% "Hierarchical Change-Detection Tests"
% IEEE Transactions on Neural Networks and Learning Systems, In Press (2016), 13 pages doi:10.1109/TNNLS.2015.2512714 
%
% Revision History
% November 2013 - merge of f_ICI_MultiChange_MultiStage and 

detected = 1;
counter = 1;
testDetectingChange = [];

p_vals = [];
h_vals = [];

DataSetReduced = DataSet;
DataSetStartIndex = 0;
confirmedDetections = [];

tsInit = 1;
tsEnd = params.training_set_length;
tsLengths = params.training_set_length;
originalDetections = tsInit;

% due to plugging-in the training set when the HT discards FP
offset = 0;

% change detection is iterated untill no change is detected (i.e. we reached the end of DataSet)
while(detected(counter) ~= 0)
    
    counter = counter + 1;
 
	params.training_set_length = tsLengths(counter - 1);
    % detect change on DataSetReduced using the trainig set identified
    % before
    %[detected(counter) , testDetectingChangeInit(counter) , Theta0 , dimensionDetected(counter)] = f_ICI_test_MultiDim(DataSetReduced, tsLengths(counter - 1) , params);
	[detected(counter)] = NPCUSUM_test(DataSetReduced, params);
    
    if detected(counter) > 0
       
	   %% prepare the datasets for validation
        originalDetections(counter) = DataSetStartIndex + detected(counter) - offset;
        
        if params.do_stop_first_detection && params.do_validate_changes == 0
            return; % the first detection has been found. Stop the CDT
        end
            
        %% hierarchical procedure, run batch hypothesis test
        % partition DataSet to perform hyopothesis test 
        if (tsInit(counter - 1) == 1)
            oldTS_Init = 1; % primo sample del DataSet, parte da quello
        else
            oldTS_Init = tsInit(counter - 1) + 1; % non siamo al primo dataset, parte da un sample dopo la detectin precedente
        end
        
        if (strcmpi(params.HT_Partition , 'TrainingSet'))
            oldTS_End = tsEnd(counter - 1);
        end
        
        if (strcmpi(params.HT_Partition , 'DataSet'))
            oldTS_End = tsInit(counter);
        end

        oldTS = DataSet(:, oldTS_Init : oldTS_End);
		
		newTS_End = originalDetections(counter);
		
		% consider a window containig the same amount of data 
		newTS_Init = originalDetections(counter) - (oldTS_End - oldTS_Init);
		% just to avoid including the same samples twice
		newTS_Init = max([newTS_Init, oldTS_End]);
		
        newTS = DataSet(:, newTS_Init : newTS_End);
		
        if (strcmpi(params.HT_Name, 'Lepage') || strcmpi(params.HT_Name, 'Mann-Whitney'))
		   DataToCheck = [oldTS, newTS];
			[h_test, p_val, T_ref] = validateChanges_CPM(DataToCheck, params.cpmParams);
		end
			
		h_vals(counter) = h_test;
		p_vals(counter) = p_val;
           
		 % CPM does not require partitioning of the data sequence to be analyzed therefore, if the change is validated
         if h_test == 1
             % cdt and ht agree: Change is Validated
             
             % we update tsInit (using \tau) provided by CPM
             tsInit(counter) = originalDetections(counter) - (size(DataToCheck, 2) - T_ref);
             % we update tsEnd as the original detection, no minimum TsLengh required
             tsEnd(counter) = originalDetections(counter);
             % we update tsEnd as the original detection, no minimum TsLengh required for validation
             tsLengths(counter) = tsEnd(counter) - tsInit(counter);
             % TODO: fixme in case the minimum TS is requred for reconfiguration of the CDT
         end
					
        %% Check Validation Results
        if h_test == 1 || (params.do_validate_changes == 0)
            
            % update DataSet and TrainigSet. (the traning set has to be in front of the valitation set)
            DataSetStartIndex = tsInit(counter);
            DataSetReduced = DataSet(: , DataSetStartIndex + 1 : end);
            offset = 0;
            
            confirmedDetections = [confirmedDetections , originalDetections(counter)];
            
            if params.do_stop_first_detection
                return;
            end
                
        else
            %Change is NOT Validated. We rely on the HT, discard detection and reconfigure detection from CDT
            detected(counter) = -1;
            
            % update DataSet and TrainigSet.
            if (tsInit(counter - 1) == 1)
                oldTS = DataSet(: , tsInit(counter - 1) : tsEnd(counter - 1));
            else
                oldTS = DataSet(: , tsInit(counter - 1) + 1 : tsEnd(counter - 1));
            end
            			
            DataSetStartIndex = originalDetections(counter); % absolute index
            % or alternatively
            % DataSetStartIndex = tsInit(counter); % questo  sbagliato! potrebbe non finire mai!!!!
            
            % con questa manovra sto scazzando i riferimenti in tsInit e tsEnd
            DataSetReduced = [oldTS , DataSet(: , DataSetStartIndex + 1 : end)];
            
            % update fields to keep track of the discarded FP
            tsInit(counter) = tsInit(counter - 1);
            tsEnd(counter) = tsEnd(counter - 1);
            tsLengths(counter) = tsLengths(counter - 1);
%             originalDetections(counter) = originalDetections(counter - 1);  % mantengo quelle originali, almeno il
%             primo output  quello senza validazione
            % compute the number of samples added to the dataSet
            offset = size(oldTS , 2);
        end
        
        if (length(DataSetReduced) <=  tsLengths(counter) ||  isempty(DataSetReduced))
            detected(counter) = 0;
            %warning('Sollevo un falso negativo perch il dataset corrisponde al ts')
        end
    end
end

% controllo se per caso non ho avuto una detection proprio all'ultimo campione per cui non posso deifinire il ts
if tsEnd(end) == length(DataSet)
    tsInit = tsInit(1 : end);
    tsEnd = tsEnd(1 : end);
    tsLengths = tsLengths(1 : end);
    testDetectingChange = testDetectingChange(1 : end - 1);
end

if exist('testDetectingChangeRefinement')
    testDetectingChangeInit(detected < 0) = - testDetectingChangeInit(detected < 0);
    testDetectingChange.Init = testDetectingChangeInit;
    testDetectingChangeRefinement(detected < 0) = - testDetectingChangeRefinement(detected < 0);
    testDetectingChange.Refined= testDetectingChangeRefinement;
end
