function [valid, tau, w, maxStat, ht, names] = runCPM(dataToBeAnalyzed, cpmParams)
%
%  [valid, tau, w, maxStat, ht, names] = runCPM(dataToBeAnalyzed, cpmParams)
% 
% tau           is the estimate of the change point given by the offline CPM. It is the partitioning maximizing the
%                   cmpParams.testType statistics
% w             is the weights of the estimate (used in  [1] to build the ensemble of CPM
% maxStat  maximum value of the cmpParams.testType statistics for all partitionings of dataToBeAnalyzed
% ht            thresholds for a given confidence \alpha to performing inference about the change-point 
% valid       validates the change, if maxStat > ht, then valid = 1 and datToBeAnalyzed contain a change-point
%                                                     if maxStat <= ht, then valid = 0 and datToBeAnalyzed does not contain a change-point
% names   of the statistics used                                                     
%
%
% References
%
% [1] GJ Ross, 
% "Parametric and Nonparametric Sequential Change Detection in R: The cpm package", 
% Journal of Statistical Software,  
%
% [2] GJ Ross, DK Tasoulis, NM Adams
% "Nonparametric Monitoring of Data Streams for Changes in Location and Scale"
% Technometrics 53 (4), 379-389
%
% [3] Cesare Alippi, Giacomo Boracchi and Manuel Roveri
% Ensembles of Change-Point Methods to Estimate the Change Point in Residual Sequences 
% Soft Computing, Springer, Volume 17, Issue 11 (2013), Page 1971-1981 doi:10.1007/s00500-013-1130-7 
%
%
% Thresholds for offline CPM are taken from  [1]
% Giacomo Boracchi
% Politecnico di Milano
% giacomo.boracchi@polimi.it 
%
% Revision Notes
% November  2013   First Release, taken from estimateT_ref_residuals(res, cpmParams)
 
TStats = zeros(cpmParams.N_TESTS, numel(dataToBeAnalyzed));
validMatrix = zeros(cpmParams.N_TESTS, numel(dataToBeAnalyzed));
htMatrix = zeros(cpmParams.N_TESTS, numel(dataToBeAnalyzed));

w = zeros(cpmParams.N_TESTS);
tau = zeros(cpmParams.N_TESTS);
maxStat = zeros(cpmParams.N_TESTS);
valid = zeros(cpmParams.N_TESTS);
ht = zeros(cpmParams.N_TESTS);

T_HAT = numel(dataToBeAnalyzed);

%%
splitPoints = [cpmParams.GAP + 1 : cpmParams.STEP : T_HAT - cpmParams.GAP];
for ss = splitPoints
    [TStats(:, ss), validMatrix(:, ss), htMatrix(:, ss), names] = computeSetDifference(dataToBeAnalyzed(1 : ss) , dataToBeAnalyzed(ss +1 : end), cpmParams);
end
% figure(643), plot(splitPoints, TStats(:,splitPoints)), title(['T statistic ' ,cpmParams.testType])
%%  determine the individual estimates and aggregation weights
for ii = 1 : cpmParams.N_TESTS
    [maxStat(ii), tau(ii)] = max(TStats(ii, :));
    valid(ii) = validMatrix(ii ,tau(ii));
    ht(ii) = htMatrix(ii, tau(ii));
    w(ii) = valid(ii) * (maxStat(ii) / ht(ii));
end


