%  This is a demo file illustrating the usage of hierarchical CDTs
%
%
%
% Please refer to the following papers, that can be downloaded from
% http://home.deib.polimi.it/boracchi/publications.html
%
% References
%
% [1] Cesare Alippi, Giacomo Boracchi, Manuel Roveri,
% Hierarchical Change-Detection Tests
% IEEE Transactions on Neural Networks and Learning Systems, In Press (2016), 13 pages doi:10.1109/TNNLS.2015.2512714
%
% [2]  Cesare Alippi, Giacomo Boracchi and Manuel Roveri,
% "A just-in-time adaptive classification system based on the intersection of confidence intervals rule,"
% Neural Networks, Elsevier vol. 24 (2011), pp. 791-800 doi: 10.1016/j.neunet.2011.05.012
%
%
% Giacomo Boracchi
% Politecnico di Milano
% March 2016
% giacomo.boracchi@polimi.it

close all
clear
clc

% randn('seed', 2)
% rand('seed', 2)

addpath(genpath('Functions/'))
addpath('SyntheticDatasets/')

%% generate dataset

AMPLITUDES= [10000, 2000]; % T^* is at time instant 2000, and then there are other 2000 samples in the new state
pdfStr = 'Gaussian';
pdfParams = [0 1; 0.5 1;]; % an abrupt change of 0.5 * \sigma to the classes mean
% pdfParams = [0 1; 0 2;]; % an abrupt change which doubles the variance
m = 0; % gradual drift parameter
do_show = 0;
do_dataset_shuffle = 0; % to be used only for classification
figureNumber = 12;
tsLength = 400;

DataSet = f_creaDataSetMultiChanges_CDT(AMPLITUDES, pdfStr , pdfParams , m , do_dataset_shuffle, do_show , figureNumber);

available_HCTDs = {'H-CUSUM', 'H-ICI+Lepage', 'H-ICI+Hotelling', 'H-NP-CUSUM'};
HCDT_toRun = available_HCTDs{2};

% % in all these HCDTs, the validation layer can be disabled by setting (copy paste these lines after initializing cdtParams)
% cdtParams.do_validate_changes = 0; %(default 1)
% % % the reconfiguration after the first detection can be disabled by setting
% cdtParams.do_stop_first_detection = 0; %(default 1)

%% run the Hierarchical CDTs (see [1] Section V.B, uncomment these lines for testing this HCDT)

switch(HCDT_toRun)
    case 'H-CUSUM'
        %% run the **Parametric** CUSUM (see [1] Section V.A, uncomment these lines for testing this HCDT)
        %set default CDT paramters
        cdtParams = define_CUSUM_standard_parameters();
        % CUSUM CDT can be tuned modifying the following paramter. Performance for different values of h are reported in [1]
        % cdtParams.h = 10;

        % distribution paramters (in CUSUM \phi_0 and \phi_1 are known)
        cdtParams.mu0 = pdfParams(1,1);
        cdtParams.sigma0 = pdfParams(1,2);
        cdtParams.mu1 = pdfParams(2,1);
        cdtParams.sigma1 = pdfParams(2,2);
        
        [tsEnd, tsInit, confirmedDetections, tsLengths, originalDetections] = f_CUSUM_MultiChange_MultiStage(DataSet, cdtParams);
        
    case 'H-ICI+Lepage'
        %% run the hierarchical ICI-based CDT using Lepage CPM at the validation layer
        cdtParams = define_ICI_test_parameters(tsLength);
        % ICI-based CDT can be tuned modifying the following paramter. Performance for different values of \Gamma are reported in [1]
        cdtParams.Gamma = 1.5;
        cdtParams.GammaRefinement = cdtParams.Gamma - 0.25; %(as in [1])
        cdtParams.do_stop_first_detection = 0;

        [tsEnd, tsInit, confirmedDetections, tsLengths, testDetectiongChange, originalDetections] = f_ICI_test_MultiChange_MultiStage(DataSet, tsLength , cdtParams);
        
    case  'H-ICI+Hotelling'
        %% run the hierarchical ICI-based CDT using the Hotelling t-test at the validation layer (see section V.B in [1])
        cdtParams = define_ICI_test_parameters(tsLength, [], [], 'Hotelling');
        % ICI-based CDT can be tuned modifying the following paramter. Performance for different values of \Gamma are reported in [1]
        % cdtParams.Gamma = 2;
        % cdtParams.GammaRefinement = cdtParams.Gamma - 0.25; (as in [1])
        
        [tsEnd, tsInit, confirmedDetections, tsLengths, testDetectiongChange, originalDetections] = f_ICI_test_MultiChange_MultiStage(DataSet, tsLength , cdtParams);
        
    case  'H-NP-CUSUM'
        %% run the NP-CUSUM Test (see [1] Section V.C, uncomment these lines for testing this HCDT)
        
        % set default CDT paramters
        cdtParams = define_NPCUSUM_standard_parameters(tsLength);
        % NP-CUSM can be tuned modifying the following paramters. Performance for different values of \kappa are reported in [1]
        % cdtParams.c.fixed.value = 0.1;
        % cdtParams.h = 450; % Threshold for the NP-CUSUM statistic (\kappa in [1])
        
        [tsEnd, tsInit, confirmedDetections, tsLengths, originalDetections] = f_NPCUSUM_MultiChange_MultiStage(DataSet, cdtParams);
        
end


%% shows the results
% each non-validated detection produces replicated entries in these vectors. For visualization purposes we remove them
tsInit = unique(tsInit);
tsEnd = unique(tsEnd);

disp([' --- '])
disp(['Change points at ', num2str(cumsum(AMPLITUDES(1 : end -1)))]);
disp(['Detections of ', ]);
disp(['Detections of the ', HCDT_toRun,' : ', num2str(unique(confirmedDetections))])
disp(['   (these detections have been confirmed by the validation layer)']);
disp(['Detections of the CDT at the first layer of the ', HCDT_toRun,' : ', num2str(unique(originalDetections))])
disp(['   (before activating the validation layer)']);
disp(['Reconfiguration of ', HCDT_toRun,' : ', num2str(unique(tsEnd(2 : end)))]);
disp(['   (including delays for gathering additional samples for validation / reconfiguration)']);
disp(['Refined estimate of the change-time instant \hat{T^*}: ', num2str(unique(tsInit(2 : end)))]);

LN_WDT = 2;
figure(1),
plot(DataSet , 'r.');
hold on
a = get(gca);
range = a.YLim;

% first detection (to set the legend entries)
plot([originalDetections(2) , originalDetections(2)] , range , 'r-' , 'LineWidth' , LN_WDT - 1);
plot([confirmedDetections(1) , confirmedDetections(1)] , range , 'b--' , 'LineWidth' , LN_WDT + 2);
plot([tsInit(2) , tsInit(2)] , range , 'g--' , 'LineWidth' , LN_WDT);
plot([tsEnd(2) , tsEnd(2)] , range , 'k:' , 'LineWidth' , LN_WDT);
% first change-point
plot([AMPLITUDES(1) , AMPLITUDES(1)] , range , 'm-' , 'LineWidth' , LN_WDT);

% by default: tsInit(1) = 1 and tsEnd(1) = tsLength. 
for ii = 2 : length(tsInit)
    plot([tsInit(ii) , tsInit(ii)] , range , 'g--' , 'LineWidth' , LN_WDT);
    plot([tsEnd(ii) , tsEnd(ii)] , range , 'k:' , 'LineWidth' , LN_WDT);
end

% next detections (\hat{T} of the first layer)
for ii = 3 : numel(originalDetections)
    plot([originalDetections(ii) , originalDetections(ii)] , range , 'r-' , 'LineWidth' , LN_WDT - 1);
end

% next detections (\hat{T} of the first layer that have been validated)
for ii = 3 : numel(confirmedDetections)
    plot([confirmedDetections(ii) , confirmedDetections(ii)] , range , 'b--' , 'LineWidth' , LN_WDT + 2);
end

% following change-points
for ii = 2 : numel(AMPLITUDES) - 1
    plot([sum(AMPLITUDES(1: ii)), sum(AMPLITUDES(1: ii))] , range , 'm-' , 'LineWidth' , LN_WDT);
end


legend('data', 'hat{T}', 'hat{T} validated', 'hat{T^*}', 'HCDT reconfiguration','True change-point (\tau)');
hold off
title(HCDT_toRun)
xlabel('time')
ylabel('observations')
axis tight




